from masks import *
#############################################################################
#
# intel hex file read, write, print, merge, etc.
#
#############################################################################

class Hex_Image:
   "read, write and manipulate Intel hex format files"

   class Hex_Image_Format_Error( IOError ):
      "errors raised by Hex_Image, derived from IOError"
      pass

   def Raise( self, Message ):
      "raise a Hex_Image_Format_Error with suitable prefix"
      raise self.Hex_Image_Format_Error, self.Error_Prefix + Message

   def __init__( self, File = None, Stride = 1, Suffixes = [ 'hex' ] ):
      "initialize, optionally specify stride and/or file to read from"
      self.Error_Prefix = ''
      self.Clear()
      self.Stride = Stride
      self.Suffixes = Suffixes
      self.Fixed = Fixed({ })
      if File <> None:
         self.Read( File )

   def Clear( self ):
      "clear all data, but do not modify stride"
      self.Data = {}
      self.Extended_Linear = 0

   def Set( self, Address, Value ):
      "set value at specified address"
      for Offset in range( 0, self.Stride ):
         self.Data[ Address * self.Stride + Offset ] = Value % 256
         Value = Value / 256;

   def Erase( self, Address ):
      for Offset in range( 0, self.Stride ):
         A = Address * self.Stride + Offset
         if self.Data.has_key( A ):
            del self.Data[ A ]

   def Get( self, Address, Default = None ):
      "return value at specified address"
      Value = 0
      for Offset in range( self.Stride - 1, -1, -1 ):
         Value = ( 256 * Value ) + \
            self.Data.get( self.Stride * Address + Offset, Default )
      if self.Fixed <> None:
         Value = self.Fixed.Fix( Address, Value )
      return Value

   def Get_Hex( self, Address, Default = None ):
      "return hex (string) representation "
      "of the value at the specified address"
      if self.Has( Address ):
         Value = self.Get( Address )
      else:
         Value = Default
      #if ( Value > 0xFF ) or ( Value < 0 ):
      #  print "%X -> %X " % ( Address, Value )
      return ( '%%0%dX' % ( 2 * self.Stride )) % Value            

   def __Has_Range( self, Base, Count ):
      "check whether all addresses in Base .. Base + Count are present"
      for Address in range( Base, Base + Count ):
         if not self.Data.has_key( Address ):
            return 0
      return 1

   def Has( self, Address ):
      "check whether Address is present, taking stride into account"
      return self.__Has_Range( Address * self.Stride, self.Stride )

   def Addresses( self ):
      "return sorted list of all addresses"
      List = self.Data.keys()
      List.sort()
      return [ A / self.Stride for A in List \
         if (( A % self.Stride == 0 ) & \
            self.__Has_Range( A, self.Stride ))]

   def __Try_Open( self, File_Name, Suffixes ):
      if Suffixes == None:
         Suffixes = self.Suffixes
      try:
         File = open( File_Name, 'r' )
         return File
      except:
         pass
      for Suffix in Suffixes:
         try:
            File = open( File_Name + '.' + Suffix, 'r' )
            return File
         except:
            pass
      File = open( File_Name, 'r' )

   def Read( self, File_Name, Suffixes = None ):
      "read from a file"
      File = self.__Try_Open( File_Name, Suffixes )
      Save_Stride = self.Stride
      self.Stride = 1
      Linear_Base = 0
      try:
         Nr = 0
         for Line in File.readlines():
            Nr = Nr + 1
            self.Error_Prefix = File_Name + ' line ' + str( Nr ) + ' '
            while ( 'x' + Line )[-1:] < ' ':
               # print "[" + Line + "]"
               Line = Line[:-1]
            if ( Line + ';' )[0] != ';': # ignore empty lines and CCS comment lines             
               if Line[0] <> ':':
                  self.Raise( 'does not start with a colon' )
               if len( Line ) % 2 <> 1:
                  self.Raise( 'does not contain an odd number of characters' )
               Length = int( Line[1:3], 16 )
               Address = int( Line[3:7], 16 )
               Type = Line[7:9]
               Checksum = ( Length + Address / 256 + \
                  Address % 256 + int( Type )) % 256
               Address = Address + Linear_Base
               Line = Line[9:]
               if Type == '00':
                  # data
                  while len(Line) > 2:
                     Data = int( Line[0:2], 16 )
                     self.Set( Address, Data )
                     Line = Line[2:]
                     Address = Address + 1
                     Checksum = ( Checksum + Data ) % 256

               elif Type == '01':
                  # end of file
                  pass

               elif Type == '02':
                  # extended segment
                   self.Raise( 'Type=2' )

               elif Type == '03':
                  # start segment
                  self.Raise( 'Type=3' )

               elif Type == '04':
                  # extended linear
                  Value = int( Line[0:4], 16 )
                  Linear_Base = 0x10000 * Value
                  Checksum = ( Checksum + Value / 256 ) + ( Value % 256 )
                  self.Extended_Linear = 1
                  Line = Line[4:]

               elif Type == '05':
                  # start linear
                  self.Raise( 'Type=5' )
   
               else:
                  self.Raise( 'unknown type=' + Type )

               if ( Checksum + int( Line, 16 )) % 256 <> 0:
                  self.Raise(
                     'checksum found=' + Line +
                     ' expected=' + ( '%02X' % (( 256 - Checksum ) % 256 )))
 
      finally:
         self.Stride = Save_Stride
         File.close()

   def Add( self, Image ):
      "add content of another image"
      for Address in Image.Data.keys:
         self.Data[ Address ] = Image.Data.get( Address )

   def __Hex_Line( self, Count, Base, Type, Data, Checksum ):
      "create data line for writing"
      Checksum = Checksum + Count + Type + \
         (( Base / 256 ) % 256 ) + ( Base % 256 )
      Checksum = ( 256 - ( Checksum % 256 )) % 256
      return ':' + \
         ( '%02X' % Count ) + \
         ( '%04X' % ( Base % 0x10000 )) + \
         ( '%02X' % Type ) + \
         Data + \
         ( '%02X' % Checksum )

   def Write( self, File_Name, Extended_Linear = None ):
      "write to a hex file"
      import os
      File_Name = os.path.normpath( File_Name )
      try:
         os.makedirs( os.path.dirname( File_Name ))
      except:
         pass
      File = open( File_Name, 'w' )
      Save_Stride = self.Stride
      self.Stride = 1
      Save_Fixed = self.Fixed
      self.Fixed = None
      try:
         if Extended_Linear == None:
            Extended_Linear = self.Extended_Linear
         if Extended_Linear:
            File.write( ':020000040000FA\n' )
         Count = 10000
         Line = ''
         Last = 0
         for Address in self.Addresses():
            if ( Count >= 16 ) \
            or ( Address <> Last + 1 ) \
            or ( Address % 16 == 0) \
            or ( Address / 0x10000 ) <> ( Last / 0x10000 ):
               if Line <> '':
                  File.write( self.__Hex_Line(
                     Count, Base, 0x00, Line, Checksum ) + '\n')
               Base = Address
               Count = 0
               Line = ''
               Checksum = 0
            if Address / 0x10000 <> Last / 0x10000:
               Segment = Address / 0x10000
               File.write( self.__Hex_Line(
                  0x02, 0x0000, 0x04,
                  ( '%04X' % Segment ),
                  Segment / 256 + Segment % 256 ) + '\n')
            Line = Line + self.Get_Hex( Address )
            Checksum = ( Checksum + self.Get( Address )) % 256
            Count = Count + 1
            Last = Address
         if Line <> '':
            File.write( self.__Hex_Line(
               Count, Base, 0x00, Line, Checksum ) + '\n')
         File.write( ':00000001FF\n' )
      finally:
         self.Stride = Save_Stride
         self.Fixed = Save_Fixed
         File.close()

   def __str__( self ):
      "return printeable representation"
      Last = -10
      Count = 0
      Result = ''
      for Address in self.Addresses():
         if ( Address <> Last + 1 ) | ( Count == ( 16 / self.Stride )):
            if Result <> '':
               Result = Result + '\n'
            Result = Result + ( '%04X:' % Address )
            Count = 0
         Result = Result + ' ' + self.Get_Hex( Address )
         Count = Count + 1
         Last = Address
      if Result == '':
         Result = 'image is empty'
      return ( 'stride=%d\n' % self.Stride ) + Result

   def String( self ):
      return self.__str__()

   def Clone( self ):
      "return cloned copy"
      import copy
      return copy.deepcopy( self )

   def Compare(
      self,
      Image,
      Name1 = 'image1',
      Name2 = 'image2',
      Range = None
   ):
      "return None or string describing first difference"
      List_1 = self.Addresses()
      List_2 = Image.Addresses()
      if self.Stride != Image.Stride:
         return "strides are different: $s=%d %s=%d" % \
            Name1, Name2, self.Stride, Image.Stride
      for Address in List_1:
         if ( Range == None ) or Address in Range:
            if len( List_2 ) == 0:
               return "address %04X in %s but not in %s" % \
                  ( Address, Name1, Name2 )
            Other = List_2.pop( 0 )
            if Address < Other:
               return "address %04X in %s but not in %s" % \
                  ( Address, Name1, Name2 )
            if Address > Other:
               return "address %04X in %s but not in %s" % \
                  ( Other, Name2, Name1 )
            if self.Get( Address ) != Image.Get( Other ):
               return "different data at address %04X: %s=%s %s=%s" % \
                  ( Address, \
                  Name1, self.Get_Hex( Address ), \
                  Name2, Image.Get_Hex( Address ) )
      return None



